#!/usr/bin/python3

from queue import Queue
from threading import Thread
import socketserver
import subprocess
from utils import log


num_gpu_tasks = 4
num_cpu_tasks = 8
HOST, PORT = "192.168.42.4", 1235

tmpdir = "/dev/shm/autoprocess"
clean_delay = 5*60 # 5 minutes

g = Queue()
c = Queue()


class DaemonTask:
    command     = None
    comm_module = None
    data        = None
    is_gpu      = False
    immediate   = False # whether the job can be ran immediately, without reserving time
    next_tasks  = []
    
    def __init__(self, command, data, is_gpu):
        try:
            self.comm_module = import_module(command)
        except ImportError as e:
            log ("Cannot load command: %s" % (str(e)))
            raise ValueError("Unable to load command %s" % command)
            
        self.command = command
        self.data    = data
        self.is_gpu  = is_gpu
        self.display = True
        try:
            self.to_run = self.comm_module.get_command(self.data)
        except Exception as e:
            log ("Cannot define task: can't build command %s with data %s: %s %s" % (self.command, self.data, type(e).__name__, e))
            raise e
            
    def run_command(self, index):
        from utils import RetryException

        log ("Spawning '{}' with data '{}', index {} on {}".format(self.to_run[0], self.data, index, "GPU" if self.is_gpu else "CPU"))
        try:
            if hasattr(self.comm_module, 'prepare'):
                self.comm_module.prepare(self.data)
            command_line = [w.replace('<index>', str(index)) for w in self.to_run]
            subprocess.check_output(command_line, stderr=subprocess.STDOUT)
        except RetryException as e:
            raise e
        except subprocess.CalledProcessError as e:
            log (e.cmd[0] + " failed: " + str(e.stdout))
        else:
            log ("Terminated '{}' with data '{}' successfully".format(self.to_run[0], self.data))
            
            # success, so enqueue the next tasks
            for task in self.next_tasks:
                enqueue(task)
    
    def set_next(self, next_tasks):
        self.next_tasks = next_tasks
    
    def describe(self, f):
        if self.display:
            import os.path
            f.write(' '.join([os.path.basename(w) for w in self.to_run]) + "\n")
        for task in self.next_tasks:
            task.describe(f)

    def __str__(self):
        return "{} on {} with data '{}'".format (self.command, "GPU" if self.is_gpu else "CPU", self.data)

def launch_task(task, index):
    from utils import RetryException
    try:
        task.run_command(index)
    except RetryException as e:
        log ("%s timed out, requeuing" % task)
        enqueue(task)
    except Exception as e:
        log ("Failed to run: " + str(e))

def worker(is_gpu, index):
    from time import sleep
    while True:
        if is_gpu:
            item = g.get()
        else:
            item = c.get()
        log ("Worker %d, %s: taking %s" % (index, "GPU" if is_gpu else "CPU", item))
        launch_task(item, index)
        if is_gpu:
            g.task_done()
            log ("{} tasks remaining on GPU".format(g.qsize()))
        else:
            c.task_done()
            log ("{} tasks remaining on CPU".format(c.qsize()))

def enqueue(task):
    if task.immediate:
        launch_task(task, 0)
    else:
        if task.is_gpu:
            g.put(task)
        else:
            c.put(task)
        log ("Queued %s" % task)

def describe_pipeline(data, task):
    import os.path

    directory = os.path.dirname(data)

    filename = os.path.join(directory, "pipeline-description.txt")

    if not os.path.isfile(filename):
        with open(filename, "w") as f:
            f.write("Examples of commands that are launched, per input file:\n")
            task.describe(f)

class DaemonTCPHandler(socketserver.StreamRequestHandler):
    def handle(self):
        for line in self.rfile.readlines():
            data = line.strip().decode('utf8')
            if data:
                import uuid
                log ("Got new job with data " + data)
                params = [data, tmpdir+'/intermediate-'+str(uuid.uuid4())]
                try:
                    task = DaemonTask("comm_clip", params, False)
                    task2 = DaemonTask("comm_motioncor2", params, True)
                    task3 = DaemonTask("comm_gctf", params, True)
                    task4 = DaemonTask("comm_remove", params[1] + '.mrc', False)
                    task4.immediate = True
                    task4.display   = False
                    task.set_next([task2])
                    task2.set_next([task3, task4])
                    enqueue(task)
                    describe_pipeline(data, task)
                except Exception as e:
                    log ("Cannot setup pipeline: %s" % (e))

def clean_tmp():
    from time import time, sleep
    from os import makedirs, listdir, path, remove

    makedirs(tmpdir, exist_ok=True)

    while True:
        time_in_secs = time() - (clean_delay)
        for i in listdir(tmpdir):
            filename = tmpdir + '/' + i
            if path.isfile(filename):
                if path.getatime(filename) <= time_in_secs:
                    # check that no queued task will use it
                    tasks = list(c.queue) + list(g.queue)
                    for name in [n for t in tasks for n in t.data]:
                        if filename == name:
                            continue
                    remove(filename)
        sleep (clean_delay)

if __name__ == "__main__":
    import socketserver
    from importlib import import_module
    import sys

    sys.dont_write_bytecode = True
    
    # spawn the consumer threads
    for i in range(num_gpu_tasks):
        t = Thread(target=worker, args=(True, i,))
        t.daemon = True
        t.start()
    for i in range(num_cpu_tasks):
        t = Thread(target=worker, args=(False, i,))
        t.daemon = True
        t.start()

    # create tmp dir if needed and spawn cleanup task
    t = Thread(target=clean_tmp)
    t.daemon = True
    t.start()
    
    # start the listening server
    server = socketserver.TCPServer((HOST, PORT), DaemonTCPHandler)
    server.serve_forever()
